/* F's IRC Bot Authentication Token */
/* $Date: 2004-03-22 23:26:59 +0100 (Mon, 22 Mar 2004) $ $Rev: 1002 $ */

/* lots of includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sha256.h"

/* our user config settings */
#include CONFIG_H

/* include a hex dump of the config file */
#include "inc-config.h"

/* some Limits */
#define MAXPATH       	260

/* Globals (yeah they suck) */

char base64_array[64] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

/* calculated sha256 of the string, then writes it in base64 into out */
/* out string buffer has 43 characters + 1 null. */
void calc_sha256(const char *str, char *out) {
    sha256_context ctx;
    uint8 digest[33];
    int m,n;

    sha256_starts(&ctx);

    sha256_update(&ctx, (const uint8*)str, strlen(str));

    sha256_finish(&ctx, digest);
    digest[32] = 0;

    n = 0;
    for(m = 0; m < 32; m += 3) {
	 
	 out[n+0] = base64_array[ (digest[m] >> 2) & 0x3F ];
	 out[n+1] = base64_array[ ((digest[m] << 4) & 0x30) + ((digest[m+1] >> 4) & 0x0F) ];
	 out[n+2] = base64_array[ ((digest[m+1] << 2) & 0x3C) + ((digest[m+2]  >> 6) & 0x03) ];
	 out[n+3] = base64_array[ (digest[m+2] & 0x3F) ];

	 n += 4;
    }
    out[n] = 0;
}

/* open file handle to (right) passwd */
FILE *open_passwd() {
     char path[MAXPATH];
     FILE *f;

     sprintf(path, "%s/%s", GLROOT, PASSWDFILE);
     
     f = fopen(path, "r");
     if (!f) {
	  /* quick dummy check so we dont attempt to read the system passwd */
	  if (getenv("RATIO") != NULL) {
	       f = fopen(PASSWDFILE, "r");
	  }	
	  if (!f) {
	       return NULL;
	  }
     }
     return f;
}

int command_showuser(const char *showuser) {
     FILE *f;
     char line[512], *token, digest[128];

     f = open_passwd();
     if (!f) {
	  printf("Error. Tell siteop we can not find the passwd file.\n");
	  return -1;
     }

     while( fgets(line, sizeof(line), f) ) {
	  char tosha[512];

	  token = strtok(line, ":");
	  if (strcmp(showuser, token) != 0) continue;

	  token = strtok(NULL, ":");

	  snprintf(tosha, sizeof(tosha), "%s %s %s", SITESALT, showuser, token);

	  calc_sha256(tosha, digest);
	  
	  printf("\n");
	  printf(" Your current invite token is %s\n", digest);
	  printf("\n");

	  fclose(f);
	  return 0;
     }

     fclose(f);

     printf("Sorry. Cannot find %s in the passwd file. You sure you exist?\n", showuser);
     return 0;
}

int command_show() {
     const char *showuser;

     showuser = getenv("USER");
     if (!showuser || !*showuser) {
	  printf("Cannot find $USER in environment. This script should be run from glftpd.\n");
	  return 0;
     }

     return command_showuser(showuser);
}

/* this loop over all users on the site. Since the sha function
   is sufficiently fast, this will not slow down anything */
int command_check(const char *authtoken) {
     FILE *f;
     char line[512], digest[128], founduser[64];
     const char *ptr;

     /* do some preliminary checks on the authtoken */
     if (strlen(authtoken) != 44) {
	  return 0;
     }

     ptr = authtoken;
     while(*ptr) {
	  if ((*ptr >= 'A' && *ptr <= 'Z') || (*ptr >= 'a' && *ptr <= 'z') ||
	      (*ptr >= '0' && *ptr <= '9') || *ptr == '+' || *ptr == '/')
	  {
	       ptr++;
	       continue;
	  }
	  
	  /* invalid character in authtoken */
	  return 0;
     }

     f = open_passwd();
     if (!f) {
	  printf("Error. Tell siteop we can not find the passwd file.\n");
	  return -1;
     }

     *founduser = 0;

     while( fgets(line, sizeof(line), f) ) {
	  char tosha[512];
	  char *this_username, *this_password;

	  this_username = strtok(line, ":");
	  this_password = strtok(NULL, ":");

	  snprintf(tosha, sizeof(tosha), "%s %s %s", SITESALT, this_username, this_password);

	  calc_sha256(tosha, digest);

	  if (strcmp(digest, authtoken) == 0) {
	       
	       if (!*founduser) {
		    strncpy(founduser, this_username, sizeof(founduser));
	       }
	       else {
		    /* we have a collusion in our user list!!! */
		    /* This shouldnt matter much thou, except if some users
		       are allowed to invite and some not. */
	       }	       
	  }
     }

     fclose(f);

     if (*founduser) {
	  printf("%s", founduser);
     }    

     return 0;
}

/* main, cuz its main */

int printhelp(void) {
     printf("Usage: f-ircpasschk <command> ...\n");
     printf("Commands:\n");
     printf("    show     	  Shows the current users token. To be used a site script.\n");
     printf("    check <token>    Verifies a user-supplied  auth token against the passwd file.\n");
     printf("                     Prints the matching user's name.\n");
     printf("    showuser <user>  Shows the specified users token.\n");
     return 0;
}

int main(int argc, char *argv[]) {

     if (argc == 2 && strcasecmp(argv[1],"dumpconfig") == 0) {
	  fwrite(confighdata, 1, sizeof confighdata, stdout);
	  return 0;
     }
     
     
     if (argc < 2) {
	  printf("Insufficient arguments for f-ircpasschk. See help.\n");
	  return 0;
     }

     if (strcasecmp(argv[1], "help") == 0) {
	  return printhelp();
     }
     else if (strcasecmp(argv[1], "show") == 0) {
	  return command_show();
     }
     else if (strcasecmp(argv[1], "showuser") == 0) {
	  if (argc != 3) {
	       printf("Syntax: f-ircpasschk showuser <user>\n");
	       return 0;
	  }

	  return command_showuser(argv[2]);
     }
     else if (strcasecmp(argv[1], "check") == 0) {
	  if (argc != 3) {
	       printf("Syntax: f-ircpasschk check <authtoken>\n");
	       return 0;
	  }

	  return command_check(argv[2]);
     }
     else {
	  printf("Unknown command for f-ircpasschk.\n");
	  return 0;
     }

     return 0;
}
